#include "catalog.h"
#include "constants.h"
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cmath>


s3d_merge::CatalogueRef::CatalogueRef(const s3d_merge::BSCLine &line)
    : m_valid(true)
    , m_bsc_id(0)
    , m_decj2000(0)
    , m_raj2000(0)
    , m_magnitude(0)
{
    const char *p(line.m_p);

    if(p)
    {
        const size_t length(strlen(p));
        char tmp[256];

        // Yale Catalogue ID.
        if(GetField(p, length, 0, 4, tmp))
        {
            m_bsc_id = atoi(tmp);
        }
        // Name
        if(GetField(p, length, 4, 10, tmp))
        {
            char *n = tmp; while(*n == ' ')n++;
            if(*n)
            {
                m_name = std::string(n);
            } 
        }
        // RA J2000
        {
            if(GetField(p, length, 75, 7, tmp))
            {
                char *n = tmp; while(*n == ' ')n++;
                if(!*n)
                {
                    m_valid = false;
                }
            }
            // Hours
            if(GetField(p, length, 75, 2, tmp))
            {
                m_raj2000 = atoi(tmp);
            }
            m_raj2000*= 60;
            // Minutes
            if(GetField(p, length, 77, 2, tmp))
            {
                m_raj2000+= atoi(tmp);
            }
            m_raj2000*= 60;
            // Seconds
            if(GetField(p, length, 79, 2, tmp))
            {
                m_raj2000+= atoi(tmp);
            }
            m_raj2000*= 10;
            // Tenths of seconds
            if(GetField(p, length, 82, 1, tmp))
            {
                m_raj2000+= atoi(tmp);
            }
        }
        // Dec J2000
        {
            if(GetField(p, length, 83, 7, tmp))
            {
                char *n = tmp; while(*n == ' ')n++;
                if(!*n)
                {
                    m_valid = false;
                }
            }

            bool positive(true);
            if(GetField(p, length, 83, 1, tmp))
            {
                if(tmp[0] == '-')
                {
                    positive = false;
                }
            }
            // degrees
            if(GetField(p, length, 84, 2, tmp))
            {
                m_decj2000 = atoi(tmp);
            }
            m_decj2000*= 60;
            // Minutes
            if(GetField(p, length, 86, 2, tmp))
            {
                m_decj2000+= atoi(tmp);
            }
            m_decj2000*= 60;
            // seconds
            if(GetField(p, length, 88, 2, tmp))
            {
                m_decj2000+= atoi(tmp);
            }
            if(!positive)
            {
                m_decj2000 = -m_decj2000;
            }
        }
        // Magnitude
        {
            bool positive(true);
            if(GetField(p, length, 102, 1, tmp))
            {
                if(tmp[0] == '-')
                {
                    positive = false;
                }
            }
            if(GetField(p, length, 103, 1, tmp))
            {
                m_magnitude = atoi(tmp);
            }
            m_magnitude*=100;
            if(GetField(p, length, 105, 2, tmp))
            {
                m_magnitude+=atoi(tmp);
            }
            if(!positive)
            {
                m_magnitude = -m_magnitude;
            }
        }
        {
            std::string ra, dec;
            GetRightAscention(ra);
            GetDeclination(dec);
        }

    }
}

double
s3d_merge::CatalogueRef::GetMagnitude(void) const 
{
    return double(m_magnitude) / 100.0;
}

double
s3d_merge::CatalogueRef::GetRelativeIntensity(void) const 
{
    double fmag = double(m_magnitude) / 100.0;

    return pow(10, fmag/-2.5);
}

//    int32_t m_decj2000; // arc seconds
//    int32_t m_raj2000;  // tenths of seconds
void
s3d_merge::CatalogueRef::GetRightAscention(std::string &out) const
{
    char buf[100];
    int32_t tmp(m_raj2000);

    int32_t hours, minutes;
    float seconds;

    seconds = float(tmp%600)/10;
    tmp/= 600;

    minutes = tmp%60;
    tmp/= 60;
    hours = tmp;

    sprintf(buf, "%i hours, %i minutes, %f seconds", hours, minutes, seconds);

    out = std::string(buf);
}

void
s3d_merge::CatalogueRef::GetDeclination(std::string &out) const
{
    char buf[100];

    int32_t tmp(abs(m_decj2000));
    int32_t degrees, minutes, seconds;

    div_t val;

    val = div(tmp, 60);
    seconds = val.rem;
    tmp = val.quot;
    
    val = div(tmp, 60);
    minutes = val.rem;
    degrees = val.quot;
    
    if(m_decj2000 < 0)
    {
        degrees = -degrees;
    }
    sprintf(buf, "%i degrees %i mins %i secs", degrees, minutes, seconds);
    out = std::string(buf);
}

bool
s3d_merge::CatalogueRef::GetField(const char *p, size_t plen, size_t fstart, size_t flen, char *out)
{
    if(p && out && ((fstart+flen) <= plen))
    {
        const char *start(p+fstart);
        strncpy(out, start, flen);
        out[flen] = 0;
        return true;
    }
    return false;
}

void
s3d_merge::CatalogueRef::Get3DVector(double *vec) const
{
    if(vec)
    {
        double dec(m_decj2000);
        double ra(m_raj2000);
        dec*= equatorial::dec_to_rad;
        ra*=  equatorial::ra_to_rad;

        printf("%g, %g\n", ra*s3d_math::one_eighty_over_pi, dec*s3d_math::one_eighty_over_pi);
        vec[0] = cos(ra) * cos(dec); // Primary direction is [1,0,0] 
        vec[1] = sin(dec);           // North celestial pole is [0,1,0]
        vec[2] = sin(ra) * cos(dec);
    }
}

void
s3d_merge::ParseYaleCatalogue(const char *fn, std::vector<CatalogueRef> &out)
{
    if(fn)
    {
        char tmp[256];
        FILE *fp;

        if((fp = fopen(fn, "r")) != NULL)
        {
            while(!feof(fp))
            {
                if(fgets(tmp, 255, fp))
                {
                    BSCLine line(tmp);
                    CatalogueRef ref(line);

                    if(ref.GetValid())
                    {
                        out.push_back(ref);
                    }
                    else
                    {
                        printf("Rejected incomplete field: \"%s\"\n", ref.GetName().c_str());
                    }
                }
            }

            fclose(fp);
        }
    }
}

