#include "image_io.h"
#include "image_types.h"
#include "merge.h"
#include "preferences.h"
#include "version.h"

#include <FreeImage.h>

#include <string>
#include <vector>

#include <cstdlib>
#include <cstdio>
#include <unistd.h>

#define DEFAULT_OUTPUT "merged.png"

void FreeImageErrorHandler(FREE_IMAGE_FORMAT fif,
                           const char *message)
{

    if(fif != FIF_UNKNOWN)
    {
        fprintf(stderr, "%s Format\n", FreeImage_GetFormatFromFIF(fif));
    }
    fprintf(stderr, "%s\n", message);
}

void show_help(void)
{
    printf("Spectral Astro Stacker %s.  Copyright (C) David Beynon 2014\n\n", SAS_VERSION);
    printf("Usage: sas_merge [options] <input1> <input2> ...\n");
    printf("Options:\n");
    printf("\t-h        : This help text.\n");
    printf("\t-p <file> : Load preferences file.\n");
    printf("\t-d        : add dark frame.  You may have many of these.\n");
    printf("\t-o <file> : Set output file.  Default is \"%s\".\n", DEFAULT_OUTPUT);
    printf("\t-t <type> : Set merge type.  Default is \"n\".\n");
    printf("\tMerge types:\n");
    printf("\t\tn : denoise/averaging.\n");
    printf("\t\ta : denoise/averaging with alignment.\n");
    printf("\t\tt : Lighten/star trails.\n");
    printf("\n%s\n", FreeImage_GetCopyrightMessage());    
}

int main(int argc, char **argv)
{
    FreeImage_Initialise(TRUE);
    FreeImage_SetOutputMessage(FreeImageErrorHandler);

    if(argc > 1)
    {
        bool align(false);
        std::vector<std::string> image_filenames;
        std::vector<std::string> dark_filenames;

        std::string output(DEFAULT_OUTPUT);

        s3d_merge::MergeType merge_type(s3d_merge::MERGE_AVERAGE);
        s3d_merge::FloatImage *output_image(NULL);
       
        s3d_merge::Preferences prefs;
 
        {
            int c;
            while((c = getopt(argc, argv, "h,d:o:p:t:")) != EOF)
            {
                switch(c)
                {
                case 'd':
                    dark_filenames.push_back(std::string(optarg));
                    break;
                case 'h':
                    show_help();
                    exit(0);
                    break;
                case 'o':
                    output = optarg;
                    fprintf(stderr, "output %s\n", optarg);
                    break;
                case 'p':
                    prefs.LoadPreferences(optarg);
                    break;
                case 't':
                    switch(*optarg)
                    {
                    case 'a':
                        align = true;
                        merge_type = s3d_merge::MERGE_AVERAGE;
                        break;
                    case 'n':
                        merge_type = s3d_merge::MERGE_AVERAGE;
                        break;
                    case 't':
                        merge_type = s3d_merge::MERGE_LIGHTEN;
                        break;
                    }
                    break;
                }
            }
            if(optind < argc)
            {
                for(int i=optind;i<argc;i++)
                {
                    image_filenames.push_back(std::string(argv[i]));
                }
            }
        }

        // Do the work.
        
        s3d_merge::FloatImage *dark_frame(NULL);

        if(!dark_filenames.empty())
        {
            printf("Averaging dark frames\n");
            dark_frame = s3d_merge::MergeNoAlign(dark_filenames, NULL, s3d_merge::MERGE_AVERAGE);
        }

        if(align)
        {
            output_image = s3d_merge::MergeAlignStars(image_filenames, dark_frame, merge_type, prefs);
        }
        else
        {
            output_image = s3d_merge::MergeNoAlign(image_filenames, dark_frame, merge_type);
        }

        if(dark_frame)
        {
            delete dark_frame;
        }

        if(output_image)
        {
            printf("Writing image \"%s\"\n", output.c_str());
            s3d_merge::WriteImageLDR(output.c_str(), output_image);

            delete output_image;
        }
    }
    else
    {
        show_help();
    }

    FreeImage_DeInitialise();

    return 0;
}
