#include "catalog.h"
#include "constants.h"
#include "match.h"
#include "align.h"
#include "image_data.h"
#include "image_io.h"
#include <stdio.h>

double
CalculateAngleThreshold(double angle)
{
     double x, y, threshold;
     angle*=(s3d_math::pi / 180.0);
     y = sin(angle);
     x = 1.0 - cos(angle);
     threshold = sqrt((x*x)+(y*y));
     return threshold;
}



class CandidateMatch
{
public:
    CandidateMatch(size_t from, size_t to)
        : m_from(from)
        , m_to(to)
        , m_score(0)
    {
    }


    size_t From(void) const
    {
        return m_from;
    }

    size_t To(void) const
    {
        return m_to;
    }

    size_t Score(void) const
    {
        return m_score;
    }

    void IncrementScore(void)
    {
        m_score++;
    }

private:
    size_t m_from,
           m_to, 
           m_score;
};

// Sort into reverse order of score so we can cull from the back end of a vector.
class CandidateMatchComparator : public std::binary_function<const CandidateMatch &, const CandidateMatch &, bool>
{
public:
    bool operator()(const CandidateMatch &a, const CandidateMatch &b)
    {
        return a.Score() > b.Score();
    }
};


int main(int argc, char ** argv)
{
    std::vector<s3d_merge::CatalogueRef> ref_stars;

    double cutoff(9);

    {
        std::vector<s3d_merge::CatalogueRef> all_stars;
        printf("Reading catalogue\n");
        s3d_merge::ParseYaleCatalogue("bsc/bsc5.dat", all_stars);

        printf("Retrieved records for %li stars\n", all_stars.size());
        s3d_merge::MagnitudeComparator cmp;

        std::sort(all_stars.begin(), all_stars.end(), cmp);

        for(size_t i=0;i<all_stars.size();i++)
        {
            if(all_stars[i].GetMagnitude() < cutoff)
            {
                ref_stars.push_back(all_stars[i]);
            }
            else
            {
                break;
            }
        }
    }

    printf("Retained %li stars with magnitude < %lf\n", ref_stars.size(), cutoff);

    std::vector<s3d_merge::Point3D> ref_points;
    std::vector<s3d_merge::Point2D> cur_points;
    ref_points.reserve(ref_stars.size());

    printf("building vector tree\n");

    for(size_t i=0;i<ref_stars.size();i++)
    {
        double tmp[3];
        ref_stars[i].Get3DVector(tmp);
        ref_points.push_back(s3d_merge::Point3D(tmp));
    }

    s3d_merge::KDTree3D ref_tree(ref_points.begin(), ref_points.end());

    printf("built vector tree\n");

    std::vector<s3d_merge::PatternNode> ref_patterns;
    std::vector<size_t> ref_pattern_indices;

    printf("building patterns\n");

    {
        double min_scale(0.05);
        double max_scale(0.075);

        while(max_scale < 10)
        {
            double tmin, tmax;

            tmin = CalculateAngleThreshold(min_scale);
            tmax = CalculateAngleThreshold(max_scale);
            printf("Scale = %g to %g\n", min_scale, max_scale);
            s3d_merge::GeneratePatternNodes3D(ref_points, &ref_tree, min_scale, max_scale, ref_patterns, ref_pattern_indices);
            min_scale*=1.5;
            max_scale*=1.5;
        }
    }

    printf("building 3D pattern tree with %li elements\n", ref_patterns.size());
    s3d_merge::PatternTree ref_pattern_tree(ref_patterns.begin(), ref_patterns.end());
    printf("Built tree\n");

    if(argc > 1)
    {
        s3d_merge::ImageData *id;

        printf("Loading image\n");
        id = s3d_merge::LoadImageData(argv[1]);

        if(id && id->Image())
        {
            std::vector<s3d_merge::StarRef> stars;
            std::vector<s3d_merge::Point2D> img_points;
            std::vector<s3d_merge::Point2D> img_orig_points;

            printf("Finding stars\n");
            s3d_merge::FindStars(id->Image(), 30, 30, 15, 10, 400, stars);
            {
                double scale; 
                s3d_merge::Point2D translate;
                translate[0] = id->Image()->get_width();
                translate[1] = id->Image()->get_height();
                translate *= 0.5;

                if(id->Image()->get_width() > id->Image()->get_height())
                {
                    scale = translate[0];
                }
                else
                {
                    scale = translate[1];
                }

                for(size_t i=0;i<stars.size();i++)
                {
                    s3d_merge::Point2D tmp;
                    tmp = ((stars[i].GetPoint() - translate) / scale);
                    tmp[1] = -tmp[1]; // coordinate hack for fitting compatibility.
                    img_points.push_back(tmp);
                    img_orig_points.push_back(stars[i].GetPoint());
                }
            }

            s3d_merge::KDTree2D img_tree(img_points.begin(), img_points.end());

            std::vector<s3d_merge::PatternNode> img_patterns;
            std::vector<size_t>                 img_pattern_indices;

            printf("Building image patterns\n");
            {
                double min_scale(0.001);
                double max_scale(0.0015);
        
                while(max_scale < 1)
                {
                    printf("Scale = %g to %g\n", min_scale, max_scale);
                    s3d_merge::GeneratePatternNodes2D(img_points, &img_tree, min_scale, max_scale, img_patterns, img_pattern_indices);

                    min_scale*=1.5;
                    max_scale*=1.5;
                }
            }

            printf("built %li img patterns\n", img_patterns.size());

            std::vector<CandidateMatch> candidate_matches;

            for(size_t i=0;i<img_patterns.size();i++)
            {
#if 0
                std::vector<size_t> found;
                found.clear();
                ref_pattern_tree.FindItemsInRange(img_patterns[i], 0.2, found);
                 
                for(size_t j=0;j<found.size();j++)
                {
                    candidate_matches.push_back(CandidateMatch(img_pattern_indices[i], ref_pattern_indices[found[j]]));
                }
#else
                size_t found;
                double distance;

                ref_pattern_tree.FindClosestItem(img_patterns[i], found, distance);

                if(distance < 0.2)
                {
                    printf("match\n");
                    candidate_matches.push_back(CandidateMatch(img_pattern_indices[i], ref_pattern_indices[found]));
                }
#endif
            }
            printf("found %li candidate matches.\n", candidate_matches.size());
            
            for(size_t i=0;i<candidate_matches.size();i++)
            {
                size_t star;
                std::string ra, dec;
                star = candidate_matches[i].To();
                ref_stars[star].GetRightAscention(ra);
                ref_stars[star].GetDeclination(dec);
                
                printf("Matched BSC ID %i, \"%s\" RA = %s, dec = %s\n", 
                       ref_stars[star].GetBSCID(), 
                       ref_stars[star].GetName().c_str(),
                       ra.c_str(), dec.c_str());
            }
            
        }
    }
    return 0;
}
